 /*
Copyright 2003 by the University of Edinburgh
Code by Dan Mossop.

SimulationCreator.java

Simulation creator is used at the start of the simulation to generate the simulation. It does so by creating instances of the templates
specified in the template file. The number of instances and their initial location is determined by the content of the object file.

*/

package code.model;

import code.model.structure.*;
import code.model.storage.*;
import code.model.utils.*;

import java.util.*;
import javax.vecmath.*;

class SimulationCreator
{
	private int numObjects = 0;
	private SimulationStore storage;
        private SimulationModel simulation;
        private ParticleDistanceStore pdStore;

	SimulationCreator(SimulationStore s, SimulationModel m, ParticleDistanceStore p)
	{
		storage = s;
                simulation = m;
                pdStore = p;
	}

        public void clear()
        {
                numObjects = 0;
        }

	////////// Register templates as part of the simulation //////////

	void addTemplate(Object template)
	{
		if (template instanceof ParticleTemplate)
		{
			// store the particle template
			storage.putTemplate(((ParticleTemplate) template).getId(), template);
		}

		if (template instanceof SimObjectTemplate)
		{
			// store the object template
			storage.putTemplate(((SimObjectTemplate) template).getId(), template);
		}

                if (template instanceof BondTemplate)
                {
                        storage.putTemplate(((BondTemplate) template).getId(), template);
                }

                if (template instanceof RuleTemplate)
                {
                        storage.putTemplate(((RuleTemplate) template).getId(), template);
                }
                if (template instanceof Landscape)
                {
                        storage.putTemplate(((Landscape) template).getId(), template);
                }
	}

	//////// Create instances of objects and return their hash strings //////////

	String createObjectInstance(SimObject obj)
	{
		return createObjectInstance(obj.getTemplateId(), "O" + (numObjects++),
						obj.getCentreOfMass(), obj.getOrientation());
	}

	private String createObjectInstance(String templateId, String objectHash,
						Vector3f comOffset, AxisAngle4f orient)
	{
		// create the object instance
		SimObjectTemplate objTemplate = (SimObjectTemplate) storage.getTemplate(templateId);
		SimObject object = new SimObject(objectHash);
		object.setSimulation(simulation);
		object.setTemplateId(templateId);

		// create the subobject instances
		String[] subObjectTemplateIds = objTemplate.getSubObjectTemplateIds();
		String[] subObjectIds = new String[subObjectTemplateIds.length];
		AxisAngle4f[] subObjOrients = objTemplate.getSubObjectOrientations();
		for (int i=0; i<subObjectTemplateIds.length; i++)
		{
			subObjectIds[i] = createObjectInstance(subObjectTemplateIds[i], objectHash + ".O" + i,
								comOffset, subObjOrients[i]);
		}
		object.setSubObjectIds(subObjectIds);

                // calculate the centre of mass of the object
                // -- compute the contribution from each of the subobjects
                Vector3f totalMassByDist = new Vector3f(); // = mass * distance
                for (int i=0; i<subObjectIds.length; i++)
                {
                        SimObject subObj = (SimObject)(storage.getInstance(subObjectIds[i]));
                        Vector3f subObjCOM = subObj.getCentreOfMass();
                        float subObjMass = subObj.getMass();
                        totalMassByDist.x += subObjMass * subObjCOM.x;
                        totalMassByDist.y += subObjMass * subObjCOM.y;
                        totalMassByDist.z += subObjMass * subObjCOM.z;
                }

		String[] particleTemplateIds = objTemplate.getParticleTemplateIds();
		String[] particleIds = new String[particleTemplateIds.length];
		Vector3f[] particleCOM = objTemplate.getParticleCentreOfMasses();

                for (int i=0; i<particleTemplateIds.length; i++)
                {
                        float particleMass = ((ParticleTemplate)
                                (simulation.getTemplate(particleTemplateIds[i]))).getMass();
                        totalMassByDist.x += particleMass * particleCOM[i].x;
                        totalMassByDist.y += particleMass * particleCOM[i].y;
                        totalMassByDist.z += particleMass * particleCOM[i].z;
                }

                Vector3f centreOfMass = new Vector3f();
                int numMasses = subObjectIds.length + particleTemplateIds.length;
                centreOfMass.x = totalMassByDist.x / numMasses;
                centreOfMass.y = totalMassByDist.y / numMasses;
                centreOfMass.z = totalMassByDist.z / numMasses;

		object.setCentreOfMass(new Vector3f(centreOfMass.x + comOffset.x,
							centreOfMass.y + comOffset.y,
							centreOfMass.z + comOffset.z));

                // get the orientations for the particles
                AxisAngle4f[] orientations = objTemplate.getParticleOrientations();

		// create the particle instances
		for (int i=0; i<particleTemplateIds.length; i++)
		{

			Vector3f newCOM = new Vector3f();
			newCOM.x = comOffset.x + particleCOM[i].x - centreOfMass.x;
			newCOM.y = comOffset.y + particleCOM[i].y - centreOfMass.x;
			newCOM.z = comOffset.z + particleCOM[i].z - centreOfMass.x;
			particleIds[i] = createParticleInstance(particleTemplateIds[i],
						objectHash + ".P" + i, newCOM, objectHash, orientations[i]);
		}
		object.setParticleIds(particleIds);

		storage.putInstance(objectHash, object);

		simulation.rotateObject(objectHash, orient);

		return objectHash;
	}

	private String createParticleInstance(String templateId, String particleHash,
						Vector3f com, String objectHash, AxisAngle4f orient)
	{
		// create the particle instance
		Particle particle = new Particle(particleHash);
		particle.setTemplateId(templateId);
		particle.setCentreOfMass(com);
		pdStore.setParticlePosition(particleHash, com);
		particle.setParentObjectId(objectHash);

		// set the bond Ids
		Vector bondIds = new Vector();
		ParticleTemplate partTemplate = (ParticleTemplate) storage.getTemplate(templateId);
		Vector bondTemplateIds = partTemplate.getBondTemplateIds();
		for (int i=0; i<bondTemplateIds.size(); i++)
		{
			String bondTemplateId = (String) (bondTemplateIds.elementAt(i));
			String bondHash = particleHash + ".B"+i;
			createBondInstance(bondTemplateId, bondHash, particleHash);
			bondIds.addElement(bondHash);
		}
		particle.setBondIds(bondIds);

		Vector3f[] bondPoints = partTemplate.getBondPoints();
		Vector3f[] transformedBondPoints =
				Functions.transformPoints(bondPoints,
					new Vector3f(), com, orient);
		particle.setBondPoints(transformedBondPoints);

		particle.setColour(partTemplate.getColour());
		particle.setAlpha(partTemplate.getAlpha());

		particle.setCollide(partTemplate.getCollide());

		particle.setOrientation(orient);

		storage.putInstance(particleHash, particle);

		return particleHash;
	}

	private String createBondInstance(String templateId, String bondHash, String parentId)
	{
		Bond bond = new Bond(bondHash, parentId, simulation);
		bond.setTemplateId(templateId);
		BondTemplate bondTemplate = (BondTemplate) (storage.getTemplate(templateId));

		// set the block rule ids
		Vector blockRuleIds = new Vector();
		Vector blockRuleTemplates = bondTemplate.getBlockRuleTemplates();
		for (int i=0; i<blockRuleTemplates.size(); i++)
		{
			String blockRuleTemplateId = (String) (blockRuleTemplates.elementAt(i));
			String blockRuleHash = bondHash + ".R" + i;
			createRuleInstance(blockRuleTemplateId, blockRuleHash);
			blockRuleIds.addElement(blockRuleHash);
		}
		bond.setBlockRuleIds(blockRuleIds);

		// set the drop rule ids
		Vector dropRuleIds = new Vector();
		Vector dropRuleTemplates = bondTemplate.getDropRuleTemplates();
		int dropOffset = blockRuleTemplates.size();
		for (int i=dropOffset; i<dropRuleTemplates.size()+dropOffset; i++)
		{
			String dropRuleTemplateId = (String) (dropRuleTemplates.elementAt(i - dropOffset));
			String dropRuleHash = bondHash + ".R" + i;
			createRuleInstance(dropRuleTemplateId, dropRuleHash);
			dropRuleIds.addElement(dropRuleHash);
		}
		bond.setDropRuleIds(dropRuleIds);

		// set the add rule ids
		Vector addRuleIds = new Vector();
		Vector addRuleTemplates = bondTemplate.getAddRuleTemplates();
		int addOffset = blockRuleTemplates.size() + dropRuleTemplates.size();
		for (int i=addOffset; i<addRuleTemplates.size()+addOffset; i++)
		{
			String addRuleTemplateId = (String) (addRuleTemplates.elementAt(i-addOffset));
			String addRuleHash = bondHash + ".R" + i;
			createRuleInstance(addRuleTemplateId, addRuleHash);
			addRuleIds.addElement(addRuleHash);
		}
		bond.setAddRuleIds(addRuleIds);

		storage.putInstance(bondHash, bond);
		return bondHash;
	}

	public String createRuleInstance(String ruleTemplateId, String ruleHash)
	{
		Rule rule = new Rule(ruleHash);
		rule.setSimulation(simulation);
		RuleTemplate ruleTemplate = (RuleTemplate) (storage.getTemplate(ruleTemplateId));

		String[] conditions = ruleTemplate.getIfClause();

		rule.setRule(ruleTemplate.getAction(), conditions, ruleTemplate.getThenClause());

                                   rule.setTemplateId(ruleTemplateId);

		storage.putInstance(ruleHash, rule);
		return ruleHash;
	}

	int getNumObjects()
	{
		return numObjects;
	}
}