 /*
Copyright 2003 by the University of Edinburgh
Code by Dan Mossop.

SimulationManipulator.java

SimulationManipulator provides a set of functions allowing easier manipulation of the data stored by the storage classes.
These methods are principly called by SimulationModel.java.

*/

package code.model;

import code.model.storage.*;
import code.model.structure.*;

import java.util.*;
import javax.vecmath.*;

class SimulationManipulator
{
        private SimulationStore storage;
        private float simulationSize, halfSimulationSize;
        private Boundary boundary;

        SimulationManipulator(SimulationStore s)
        {
                storage = s;
                boundary = new Boundary();
        }

        public void clear()
        {
                storage.clear();
        }

        // -- boundary operations
        public Vector3f getPosition(Vector3f pos)
        {
                return boundary.getPosition(pos);
        }

        public boolean testPosition(Vector3f pos)
        {
                return boundary.testPosition(pos);
        }

        public void setSimulationSize(float s)
        {
                boundary.setSimulationSize(s);
        }
		
		public float getSimulationSize()
		{
				return boundary.getSimulationSize();
		}

        public void setBoundaryType(String type)
        {
                boundary.setType(type);
        }

        public float getSquaredDistance(Vector3f vect1, Vector3f vect2)
        {
                return boundary.getSquaredDistance(vect1, vect2);
        }

        public float getDistance(Vector3f vect1, Vector3f vect2)
        {
                return boundary.getDistance(vect1, vect2);
        }
		
		public Vector3f getVectorBetween(Vector3f vect1, Vector3f vect2)
		{
				return boundary.getVectorBetween(vect1, vect2);
		}

        // -- single particle operations

        public float getParticleRadius(String id)
        {
                Particle p = (Particle)(storage.getInstance(id));
                ParticleTemplate pt = (ParticleTemplate)(storage.getTemplate(p.getTemplateId()));
                return pt.getShapeRadius();
        }
		
        public float getParticleVol(String id)
        {
                Particle p = (Particle)(storage.getInstance(id));
                ParticleTemplate pt = (ParticleTemplate)(storage.getTemplate(p.getTemplateId()));
                return pt.getShapeVol();
        }

        public float getParticleMass(String id)
        {
                Particle p = (Particle)(storage.getInstance(id));
                ParticleTemplate pt = (ParticleTemplate)(storage.getTemplate(p.getTemplateId()));
                return pt.getMass();
        }
		
	public Landscape getParticleLandscape(String id)
	{
			Particle p = (Particle)(storage.getInstance(id));
			ParticleTemplate pt = (ParticleTemplate)(storage.getTemplate(p.getTemplateId()));
			return (Landscape) (storage.getTemplate(pt.getLandscapeId()));
	}

	public Vector getParticleBondIds(String id)
	{
			Particle p = (Particle)(storage.getInstance(id));
			return p.getBondIds();
	}

	public Vector3f[] getParticleBondPoints(String id)
	{
			Particle p = (Particle)(storage.getInstance(id));
			return p.getBondPoints();
	}

	public void setParticleBondPoints(String id, Vector3f[] points)
	{
			Particle p = (Particle)(storage.getInstance(id));
			p.setBondPoints(points);
	}

        /*
        public Vector3f translateParticle(String id, Vector3f trans)
        {
                Particle p = (Particle)(storage.getInstance(id));

                Vector3f com = p.getCentreOfMass();
                com.x += trans.x;
                com.y += trans.y;
                com.z += trans.z;
                Vector3f newPos = boundary.getPosition(com);
                p.setCentreOfMass(newPos);

                return newPos;
        }
        */

        public Vector3f translatePoint(Vector3f com, Vector3f trans)
        {
                com.x += trans.x;
                com.y += trans.y;
                com.z += trans.z;
                Vector3f newPos = boundary.getPosition(com);
                return newPos;
        }

        public Vector3f getParticleCentreOfMass(String id)
        {
                Particle p = (Particle)(storage.getInstance(id));
                return p.getCentreOfMass();
        }

        public void setParticleCentreOfMass(String id, Vector3f com)
        {
                Particle p = (Particle)(storage.getInstance(id));
                p.setCentreOfMass(com);
        }
		
	public void setParticleColour(String id, Color3f col)
	{
			Particle p = (Particle)(storage.getInstance(id));
			p.setColour(col);
	}

        public AxisAngle4f getParticleOrientation(String id)
        {
                Particle p = (Particle)(storage.getInstance(id));
                return p.getOrientation();
        }

        public void setParticleOrientation(String id, AxisAngle4f orient)
        {
                Particle p = (Particle)(storage.getInstance(id));
                p.setOrientation(orient);
        }

        // -- multiple particle operations

        String[] getAllParticlesOfObject(String objId)
        {
                Vector vect = new Vector();
                getParticleIdsRecursiveVect(vect, objId);
                String[] ids = new String[vect.size()];
                for (int i=0; i<vect.size(); i++)
                {
                        ids[i] = (String) (vect.elementAt(i));
                }
                return ids;
        }

        public Vector getParticleIdsRecursiveVect(Vector vect, String objId)
        {
                SimObject obj = (SimObject) (storage.getInstance(objId));
                if (obj == null)
                {
                        return vect;
                }

                String[] particleIds = obj.getParticleIds();
                for (int i=0; i<particleIds.length; i++)
                {
                        vect.addElement(particleIds[i]);
                }

                String[] subObjectIds = obj.getSubObjectIds();
                for (int i=0; i<subObjectIds.length; i++)
                {
                        getParticleIdsRecursiveVect(vect, subObjectIds[i]);
                }
                return vect;
        }

        String[] getAllRuleTemplates()
        {
                Set set = storage.getTemplateEntrySet();
                Vector rules = new Vector();

                Iterator iter = set.iterator();
                for (int i=0; i<set.size(); i++)
                {
                        Map.Entry me = (Map.Entry) iter.next();

                        Object obj = me.getValue();

                        if (obj instanceof RuleTemplate)
                        {
                                rules.addElement(me.getKey());
                        }
                }

                String[] ruleIds = new String[rules.size()];
                for (int i=0; i<rules.size(); i++)
                {
                        ruleIds[i] = (String) (rules.elementAt(i));
                }

                return ruleIds;
        }

        String[] getAllParticleIds()
        {
                Set set = storage.getInstanceEntrySet();
                Vector particles = new Vector();

                Iterator iter = set.iterator();
                for (int i=0; i<set.size(); i++)
                {
                        Map.Entry me = (Map.Entry) iter.next();

                        Object obj = me.getValue();

                        if (obj instanceof Particle)
                        {
                                particles.addElement(me.getKey());
                        }
                }

                String[] partIds = new String[particles.size()];
                for (int i=0; i<particles.size(); i++)
                {
                        partIds[i] = (String) (particles.elementAt(i));
                }

                return partIds;
        }

        // -- multiple object operations

        String[] getAllObjectIds()
        {
                Set set = storage.getInstanceEntrySet();
                Vector objects = new Vector();

                Iterator iter = set.iterator();
                for (int i=0; i<set.size(); i++)
                {
                        Map.Entry me = (Map.Entry) iter.next();

                        Object obj = me.getValue();

                        if (obj instanceof SimObject)
                        {
                                objects.addElement(me.getKey());
                        }
                }

                String[] objIds = new String[objects.size()];
                for (int i=0; i<objects.size(); i++)
                {
                        objIds[i] = (String) (objects.elementAt(i));
                }

                return objIds;
        }

        Vector getAllTemplates()
        {
                Set set = storage.getTemplateEntrySet();
                Vector objectTemplates = new Vector();

                Iterator iter = set.iterator();
                for (int i=0; i<set.size(); i++)
                {
                        Map.Entry me = (Map.Entry) iter.next();
                        Object obj = me.getValue();
                        objectTemplates.addElement(obj);
                }

                return objectTemplates;
        }
}

class Boundary
{
        private String type = "walled";
        private float simulationSize, halfSimulationSize;

        Boundary()
        {
                setSimulationSize(1f);
        }

        void setSimulationSize(float simSize)
        {
                simulationSize = simSize;
                // -- compute half of sim size for efficiency
                halfSimulationSize = 0.5f * simSize;
        }
		
		float getSimulationSize()
		{
			return simulationSize;
		}

        void setType(String t)
        {
                type = t;
        }

        boolean testPosition(Vector3f c)
        {
                // switch for different boundaries
                if (type.equals("tesselated"))
                {
                        return true;
                }
                else
                {
                        // assume boundary = "walled"
                        return testPositionWalled(c);
                }
        }

        boolean testPositionWalled(Vector3f c)
        {
                if (c.x >= halfSimulationSize || c.x <= -halfSimulationSize) { return false; }
                if (c.y >= halfSimulationSize || c.y <= -halfSimulationSize) { return false; }
                if (c.z >= halfSimulationSize || c.z <= -halfSimulationSize) { return false; }
                return true;
        }

        Vector3f getPosition(Vector3f c)
        {
                // switch for different boundaries
                if (type.equals("tesselated"))
                {
                        return getPositionTesselated(c);
                }
                else
                {
                        // assume boundary = "walled"
                        return c;
                }
        }

        Vector3f getPositionTesselated(Vector3f c)
        {
                Vector3f com = new Vector3f(c);

                while (com.x >= halfSimulationSize) {com.x -= simulationSize;}
                while (com.x <= -halfSimulationSize){com.x += simulationSize;}
                while (com.y >= halfSimulationSize) {com.y -= simulationSize;}
                while (com.y <= -halfSimulationSize){com.y += simulationSize;}
                while (com.z >= halfSimulationSize) {com.z -= simulationSize;}
                while (com.z <= -halfSimulationSize){com.z += simulationSize;}

                return com;
        }
		
		Vector3f getVectorBetween(Vector3f v1, Vector3f v2)
		{
                Vector3f v3 = new Vector3f(v2.x - v1.x, v2.y - v1.y, v2.z - v1.z);
				return v3;
		}

        float getDistance(Vector3f v1, Vector3f v2)
        {
                // switch for different boundaries if present - not at the moment
				if (type.equals("tesselated"))
				{
                	return getDistanceTesselated(v1, v2);
        		}
				else
				{
					return getDistanceWalled(v1, v2);
				}
		}
		
		float getDistanceWalled(Vector3f v1, Vector3f v2)
		{
				return (float) Math.sqrt( getSquaredDistanceWalled(v1, v2) );
		}
		
		float getSquaredDistanceWalled(Vector3f v1, Vector3f v2)
        {
                Vector3f v3 = new Vector3f(v2.x - v1.x, v2.y - v1.y, v2.z - v1.z);
                return (v3.x*v3.x + v3.y*v3.y + v3.z*v3.z);
        }

        float getDistanceTesselated(Vector3f v1, Vector3f v2)
        {
                return (float) Math.sqrt( getSquaredDistanceTesselated(v1, v2) );
        }

        float getSquaredDistance(Vector3f v1, Vector3f v2)
        {
                // switch for different boundaries if present - not at the moment
				if (type.equals("tesselated"))
				{
                	return getSquaredDistanceTesselated(v1, v2);
        		}
				else
				{
					return getSquaredDistanceWalled(v1, v2);
				}
		}

        float getSquaredDistanceTesselated(Vector3f v1, Vector3f v2)
        {
                Vector3f v3 = new Vector3f(v2.x - v1.x, v2.y - v1.y, v2.z - v1.z);
                v3 = getPositionTesselated(v3);
                return (v3.x*v3.x + v3.y*v3.y + v3.z*v3.z);
        }

}





