 /*
Copyright 2003 by the University of Edinburgh
Code by Dan Mossop.

SimulationModel.java

SimulationModel provides the main methods called by Evolver.java. It brings together the simulation creator and manipulator,
the file handler and the display, allowing them to be controlled through single method calls. It is responsible for initialising
all of the above, as well as the UpdatingCanvas3D object (defined at the end of this file) which wraps the Java3d renderer,
allowing the timing of frame advances to be controlled.

*/

package code.model;

import gui.*;
import code.model.storage.*;
import code.model.display.*;
import code.model.structure.*;
import code.model.reader.*;
import code.model.utils.*;
import code.update.*;

import java.awt.event.*;
import java.awt.*;
import java.io.*;

import javax.swing.event.*;
import javax.swing.*;

import javax.vecmath.*;
import java.util.*;

import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.behaviors.mouse.*;
import javax.media.j3d.*;
import javax.vecmath.*;

public class SimulationModel
{
        ParticleDistanceStore pdStore;
        SimulationStore storage;
        SimulationCreator creator;
        SimulationManipulator manipulator;
        FileHandler fileHandler;

        Display display;
        UpdatingCanvas3D uCanvas3d;


        public SimulationModel(Point loc, Dimension dim)
        {
                pdStore = new ParticleDistanceStore();
                storage = new SimulationStore();
                creator = new SimulationCreator(storage, this, pdStore);
                manipulator = new SimulationManipulator(storage);
                uCanvas3d = new UpdatingCanvas3D();
                display = new Display(uCanvas3d, loc, dim);
                uCanvas3d.startRenderer();
                fileHandler = new FileHandler();
        }

        public void dispose()
        {
                uCanvas3d.stopRenderer();
                display.dispose();
        }

        public Point getLocation()
        {
                return display.getLocation();
        }

        public Dimension getDimensions()
        {
                return display.getDimensions();
        }
		
        public void setEvolver(Evolver e)
        {
                uCanvas3d.setEvolver(e);
        }

        public void setRunning(boolean r)
        {
                uCanvas3d.setRunning(r);
        }

        public ParticleDistanceStore getPDStore()
        {
                return pdStore;
        }
		
        public SimulationStore getStorage()
        {
                return storage;
        }

        public void addTemplate(Object o)
        {
                creator.addTemplate(o);
        }

        public void createObjectInstance(SimObject object)
        {
                creator.createObjectInstance(object);
        }

        public void putInstance(String hashValue, Object object)
        {
                storage.putInstance(hashValue, object);
        }

        public Object getInstance(String hashValue)
        {
                return storage.getInstance(hashValue);
        }

        public void putTemplate(String hashValue, Object object)
        {
                storage.putTemplate(hashValue, object);
        }

        public Object getTemplate(String hashValue)
        {
                return storage.getTemplate(hashValue);
        }

        public void printInstanceHashMap()
        {
                storage.printInstanceHashMap();
        }

        public void printTemplateHashMap()
        {
                storage.printTemplateHashMap();
        }

        //////////////////////////////////////////////////////////////////

        public String[] getAllParticles()
        {
                return manipulator.getAllParticleIds();
        }

        public String[] getAllObjects()
        {
                return manipulator.getAllObjectIds();
        }

        public String[] getAllRuleTemplates()
        {
                return manipulator.getAllRuleTemplates();
        }
		
        public void clear()
        {
                storage.clear();
                pdStore.clear();
                creator.clear();
                manipulator.clear();
                display();
        }


        public void display()
        {
                String[] particles = getAllParticles();
                display.initialiseParticles(particles);

                for (int i=0; i<particles.length; i++)
                {
                        Particle p = (Particle) (storage.getInstance(particles[i]));
                        ParticleTemplate pt = (ParticleTemplate) (storage.getTemplate(p.getTemplateId()));

                        // set the particle shape and colour
                        display.setPrimitive(particles[i], pt.getShapeName(), pt.getShapeDimensions());
                        display.setColour(particles[i], p.getColour());
						display.setAlpha(particles[i], p.getAlpha());

                        // set the centre of mass
                        display.setCentreOfMass(particles[i], p.getCentreOfMass());
			display.setOrientation(particles[i], p.getOrientation());
                }

                display.showParticles(particles);
        }

        public void setSimulationSize(float s)
        {
                manipulator.setSimulationSize(s);
                display.setScale(10f / s);
        }
		
		public float getSimulationSize()
		{
			return manipulator.getSimulationSize();
		}

        // File handling
        public void writeTemplateFile(String filename)
        {
                fileHandler.toFile(filename, manipulator.getAllTemplates());
        }

        public void loadFile(String filename)
        {
                Vector objects = fileHandler.fromFile(filename);

                for (int i=0; i<objects.size(); i++)
                {
                        Object o = (objects.elementAt(i));
                        
                        if (o instanceof SimObject)
                        {
                                createObjectInstance((SimObject) o);
                        }
                        else
                        {
                                        addTemplate(o);
                        }
                }
        }

        //////////////////////////////////////////////////////////
        // -- user methods (ie. available to the update algorithm)

        public String[] getRootObjects()
        {
                int num = creator.getNumObjects();
                String[] hashes = new String[num];
                for (int i=0; i<num; i++)
                {
                        hashes[i] = "O"+i;
                }
                return hashes;
        }

        public String getRootObject(int num)
        {
                if (num < creator.getNumObjects())
                {
                        return "O"+num;
                }
                return "";
        }

        // -- operations on a single particle

        public float getParticleRadius(String id)
        {
                return manipulator.getParticleRadius(id);
        }
		
        public float getParticleVol(String id)
        {
                return manipulator.getParticleVol(id);
        }

        public void translateParticle(String id, Vector3f trans)
        {
	Vector3f oldCom = getParticleCentreOfMass(id);
                Vector3f com = manipulator.translatePoint(oldCom, trans);
                setParticleCentreOfMass(id, com);
                display.setCentreOfMass(id, com);
                pdStore.setParticlePosition(id, com);

                // update the particle bonds
                Vector3f[] oldBondPoints = getParticleBondPoints(id);
                Vector3f[] newBondPoints = Functions.transformPoints(oldBondPoints, oldCom,
                                                                                                trans, new AxisAngle4f());
                setParticleBondPoints(id, newBondPoints);
        }

        public Vector3f translatePoint(Vector3f com, Vector3f trans)
        {
                return manipulator.translatePoint(com, trans);
        }

	public void rotateParticle(String particleId, Vector3f objectCOM, AxisAngle4f orient)
	{
                                   AxisAngle4f partOrient = getParticleOrientation(particleId);

		AxisAngle4f newOrient = Functions.combine(partOrient, orient);
 		if (newOrient == null)
                                        {
                                                return;
                                        }

                                        setParticleOrientation(particleId, newOrient);

                                        // update the particle bonds
                                        Vector3f[] oldBondPoints = getParticleBondPoints(particleId);
                                        Vector3f[] newBondPoints = Functions.transformPoints(oldBondPoints, objectCOM, new Vector3f(), orient);
                                        setParticleBondPoints(particleId, newBondPoints);
	}

	public void rotateObject(String id, AxisAngle4f orient)
	{
			SimObject obj = (SimObject) (getInstance(id));
			Vector3f objectCOM = obj.getCentreOfMass();
			Vector particleIds = obj.getAllParticleIds();
			rotateObject(particleIds, objectCOM, orient);
	}

	boolean rotateObject(Vector particleIds, Vector3f objectCOM, AxisAngle4f orient)
	{
		AxisAngle4f[] newOrients = new AxisAngle4f[particleIds.size()];
		for (int i=0; i<particleIds.size(); i++)
		{
			String particleId = (String) (particleIds.elementAt(i));

			AxisAngle4f partOrient = getParticleOrientation(particleId);

			newOrients[i] = Functions.combine(partOrient, orient);
 			if (newOrients[i] == null)
 			{
 				return false;
			}
		}

                                  Vector3f[] points = new Vector3f[particleIds.size()];
                                  for (int i=0; i<particleIds.size(); i++)
                                  {
                                        String particleId = (String) (particleIds.elementAt(i));
                                        points[i] = getParticleCentreOfMass(particleId);
                                  }

                                  Vector3f[] finalPos = Functions.transformPoints(points, objectCOM, new Vector3f(), orient);

		for (int i=0; i<particleIds.size(); i++)
		{
			String particleId = (String) (particleIds.elementAt(i));

  			setParticleOrientation(particleId, newOrients[i]);
			setParticleCentreOfMass(particleId, finalPos[i]);

			// update the particle bonds
			Vector3f[] oldBondPoints = getParticleBondPoints(particleId);
			Vector3f[] newBondPoints = Functions.transformPoints(oldBondPoints, objectCOM, new Vector3f(), orient);
			setParticleBondPoints(particleId, newBondPoints);
		}
		return true;
	}


        public Vector3f getParticleCentreOfMass(String id)
        {
                return manipulator.getParticleCentreOfMass(id);
        }

        public void setParticleCentreOfMass(String id, Vector3f com)
        {
		pdStore.setParticlePosition(id, com);
                manipulator.setParticleCentreOfMass(id, com);
                display.setCentreOfMass(id, com);
        }
		
	public Landscape getParticleLandscape(String id)
	{
			return manipulator.getParticleLandscape(id);
	}

	public Vector getParticleBondIds(String id)
	{
			return manipulator.getParticleBondIds(id);
	}

	public Vector3f[] getParticleBondPoints(String id)
	{
			return manipulator.getParticleBondPoints(id);
	}

	public void setParticleBondPoints(String id, Vector3f[] bp)
	{
			manipulator.setParticleBondPoints(id, bp);
	}

	public String getParticleTemplate(String id)
	{
		Particle p = (Particle) storage.getInstance(id);
		return p.getTemplateId();
	}

	public void setParticleColour(String id, Color3f col)
	{
			manipulator.setParticleColour(id, col);
			display.setColour(id, col);
	}

        public AxisAngle4f getParticleOrientation(String id)
        {
                return manipulator.getParticleOrientation(id);
        }

        public void setParticleOrientation(String id, AxisAngle4f orient)
        {
                manipulator.setParticleOrientation(id, orient);
                display.setOrientation(id, orient);
        }

        public float getParticleMass(String id)
        {
                return manipulator.getParticleMass(id);
        }

        // -- operations on multiple particles
        public String[] getAllParticlesOfObject(String id)
        {
                return manipulator.getAllParticlesOfObject(id);
        }

        // -- boundary operations
        public Vector3f getPosition(Vector3f pos)
        {
                return manipulator.getPosition(pos);
        }

        public boolean testPosition(Vector3f pos)
        {
                return manipulator.testPosition(pos);
        }

        public float getSquaredDistance(Vector3f vect1, Vector3f vect2)
        {
                return manipulator.getSquaredDistance(vect1, vect2);
        }

        public float getDistance(Vector3f vect1, Vector3f vect2)
        {
                return manipulator.getDistance(vect1, vect2);
        }
		
        public Vector3f getVectorBetween(Vector3f vect1, Vector3f vect2)
        {
                return manipulator.getVectorBetween(vect1, vect2);
        }

        public void setFrameDelay(int fd)
        {
                if (fd >= 0)
                {
                        uCanvas3d.setFrameDelay(fd);
                }
        }

        public int getFrameDelay()
        {
                return uCanvas3d.getFrameDelay();
        }
}

class UpdatingCanvas3D extends Canvas3D
{
	boolean isRunning = false;
	Evolver evolver;
	final boolean DEBUG = false;
                //long start;
                //long end;
                int frameDelay = 0;

	UpdatingCanvas3D()
	{
		super((new GraphConfig()).getGraphConfig());
	}

	public void setEvolver(Evolver e)
	{
		evolver = e;
	}

	public void preRender()
	{
		if (isRunning && evolver != null)
		{
			evolver.update();
		}
                                else
                                {
                                        try
                                        {
                                                Thread.currentThread().sleep(100);
                                        }
                                        catch(Exception e)
                                        {
                                        }
                                }
                        //start = (new Date()).getTime();
	}

	public void postSwap()
	{
                        //end  = (new Date()).getTime();
                        //System.out.println("render : "+(end-start)+" ms");

	        if (isRunning)
                         {
                                sim.updateElapsedTime();
                        }

                        try
                        {
                                Thread.currentThread().sleep(frameDelay);
                        }
                        catch(Exception e)
                        {
                        }

		Graphics g = this.getGraphics();
		this.paint(g);
	}

	void setRunning(boolean r)
	{
		isRunning = r;
	}

                public void setFrameDelay(int fd)
                {
                        frameDelay = fd;
                }

                public int getFrameDelay()
                {
                        return frameDelay;
                }
}

class GraphConfig
{
	GraphicsConfiguration getGraphConfig()
	{
		GraphicsConfigTemplate3D tmpl = new GraphicsConfigTemplate3D( );
		GraphicsEnvironment env =
				GraphicsEnvironment.getLocalGraphicsEnvironment( );
		GraphicsDevice device = env.getDefaultScreenDevice( );
		GraphicsConfiguration config = device.getBestConfiguration( tmpl );
		return config;
	}
};