  /*
Copyright 2003 by the University of Edinburgh
Code by Dan Mossop.

ParticleDistanceStore.java

Implements a data structure used to store the positions of particles in three dimensions.
This allows rapid retrieval of all the particles in a given area, speeding up collision detection and bond testing. 

*/

package code.model.storage;

import java.util.*;
import javax.vecmath.*;

public class ParticleDistanceStore
{
        private Vector xcoord;
		private Vector xparticle;
		private Vector ycoord;
		private Vector yparticle;
		private Vector zcoord;
		private Vector zparticle;

        public ParticleDistanceStore()
        {
			xcoord = new Vector();
			xparticle = new Vector();
			ycoord = new Vector();
			yparticle = new Vector();
			zcoord = new Vector();
			zparticle = new Vector();
        }
		
		public void clear()
		{
			xcoord = new Vector();
			xparticle = new Vector();
			ycoord = new Vector();
			yparticle = new Vector();
			zcoord = new Vector();
			zparticle = new Vector();
		}
		
		public Vector getAllParticles()
		{
			return new Vector(xparticle);
		}
		
		public Vector getNearbyParticles(Vector3f pos, float axialDist)
		{
			Vector xnearby = particlesInRange(xcoord, xparticle, pos.x-axialDist, pos.x+axialDist);
			Vector ynearby = particlesInRange(ycoord, yparticle, pos.y-axialDist, pos.y+axialDist); 
			Vector znearby = particlesInRange(zcoord, zparticle, pos.z-axialDist, pos.z+axialDist); 
			
			Vector nearbyParticles = new Vector();
			for (int i=0; i<xnearby.size(); i++)
			{
				Object part = xnearby.elementAt(i);
				if (ynearby.indexOf(part) != -1 && znearby.indexOf(part) != -1)
				{
					nearbyParticles.addElement(part);
				}
			}
			return nearbyParticles;
		}
		
		public Vector particlesInRange(Vector coord, Vector particle, float minPos, float maxPos)
		{
			int start = minIndex(coord, minPos, 0, coord.size());
			int end = maxIndex(coord, maxPos, 0, coord.size());
			
			Vector partsInRange = new Vector();
			for (int i=start; i<end; i++)
			{
				partsInRange.addElement(particle.elementAt(i));
			}
			return partsInRange;
		}
		
		public int minIndex(Vector coord, float pos, int min, int max)
		{
			// -- greater than one element in the range
			if (max-min > 1)
			{
				int split = (min+max) / 2;
				if (pos <= ((Float) (coord.elementAt(split))).floatValue())
				{
					return minIndex(coord, pos, min, split);
				}
				else
				{
					return minIndex(coord, pos, split+1, max);
				}
			}
			
			// -- zero elements in the range			
			if (min == max)
			{
				return min;
			}
			
			// -- one element in the range
			if (pos <= ((Float) (coord.elementAt(min))).floatValue())
			{
				return min;
			} 
			else
			{
				return max;
			}			
		}
		
		public int maxIndex(Vector coord, float pos, int min, int max)
		{
			// -- greater than one element in the range
			if (max-min > 1)
			{
				int split = (min+max) / 2;
				if (pos < ((Float) (coord.elementAt(split))).floatValue())
				{
					return maxIndex(coord, pos, min, split);
				}
				else
				{
					return maxIndex(coord, pos, split+1, max);
				}
			}
			
			// -- zero elements in the range			
			if (min == max)
			{
				return min;
			}
			
			// -- one element in the range
			if (pos < ((Float) (coord.elementAt(min))).floatValue())
			{
				return min;
			} 
			else
			{
				return max;
			}			
		}
		
		public void setParticlePosition(String id, Vector3f position)
		{
			// -- remove the particle if it is already in the store
			removeFrom(xcoord, xparticle, id);
			removeFrom(ycoord, yparticle, id);
			removeFrom(zcoord, zparticle, id);
			
			// -- add it in the new position
			insertInto(xcoord, xparticle, 0, xcoord.size(), id, position.x);
			insertInto(ycoord, yparticle, 0, ycoord.size(), id, position.y);
			insertInto(zcoord, zparticle, 0, zcoord.size(), id, position.z);
		}
		
		public void removeFrom(Vector coord, Vector particle, String id)
		{
			int index = particle.indexOf(id);
			if (index != -1)
			{
				coord.removeElementAt(index);
				particle.removeElementAt(index);
			}
		}
		
		public void insertInto(Vector coord, Vector particle, int min, int max, String id, float position)
		{
			// -- greater than one element in the range
			if (max-min > 1)
			{
				int split = (min+max) / 2;
				if (position <= ((Float) (coord.elementAt(split))).floatValue())
				{
					insertInto(coord, particle, min, split, id, position);
					return;
				}
				else
				{
					insertInto(coord, particle, split+1, max, id, position);
					return;
				}
			}
			
			// -- zero elements in the range			
			if (min == max)
			{
				coord.insertElementAt(new Float(position), min);
				particle.insertElementAt(id, min);
				return;
			}
			
			// -- one element in the range
			if (position <= ((Float) (coord.elementAt(min))).floatValue())
			{
				coord.insertElementAt(new Float(position), min);
				particle.insertElementAt(id, min);
			} 
			else
			{
				coord.insertElementAt(new Float(position), max);
				particle.insertElementAt(id, max);
			}
		}
		
		public void print()
		{
			System.out.println("PARTICLE DISTANCE STORE\n\nx coordinates\n");
			for (int i=0; i<xcoord.size(); i++)
			{
				System.out.println( ((Float)(xcoord.elementAt(i))).floatValue() +
				" -> " + ((String)(xparticle.elementAt(i))));
			}
			System.out.println("\ny coordinates\n");
			for (int i=0; i<ycoord.size(); i++)
			{
				System.out.println( ((Float)(ycoord.elementAt(i))).floatValue() +
				" -> " + ((String)(yparticle.elementAt(i))));
			}
			System.out.println("\nz coordinates\n");
			for (int i=0; i<zcoord.size(); i++)
			{
				System.out.println( ((Float)(zcoord.elementAt(i))).floatValue() +
				" -> " + ((String)(zparticle.elementAt(i))));
			}
		}
		
		public static void main(String[] args)
		{
			ParticleDistanceStore store = new ParticleDistanceStore();
			
			store.setParticlePosition("p1", new Vector3f(0f,0f,0f));
			store.print();
			store.setParticlePosition("p1", new Vector3f(1f, 0f, 0f));
			store.print();
			store.setParticlePosition("p2", new Vector3f(2f, 0f, 0f));
			store.print();
			store.setParticlePosition("p3", new Vector3f(0f, 0f, 0f));
			store.print();
			store.setParticlePosition("p4", new Vector3f(2f, 1f, -1f));
			store.print();
			Vector nearby = store.getNearbyParticles(new Vector3f(0f, 0f, 0f), 1.75f);
			for (int i=0; i<nearby.size(); i++)
			{
				System.out.println((String)(nearby.elementAt(i)));
			}
		}
}