 /*
Copyright 2003 by the University of Edinburgh
Code by Dan Mossop.

Evolver.java

Defines the Evolver object which advances the simulation at each timestep. This file is responsible for calculating the new positions of objects.
It also coordinates the process of checking for changes in the bond arrangements and updates the particle positions to fulfill these bonds.

*/

package code.update;

import code.model.SimulationModel;
import code.model.structure.*;
import code.model.utils.*;

import java.util.*;
import javax.vecmath.*;

public class Evolver
{
        SimulationModel simulation;

        final float BOLTZMANN_CONSTANT = 1.3807f * (float) Math.pow(10.0, -23.0);

        float diffusionCoefficient;
        Random rand = new Random();
        float timestep = 0f;
        float waterViscosity = 0f;
        float temperature = 300f;

        float acceptanceFactor = 0f;

        int totalChanges = 0;
        int acceptedChanges = 0;

        private Vector clusteredParticles = new Vector();
        private Vector clusteredIds = new Vector();

        private Vector clusterLookup = new Vector();
        private Vector clusterCentresOfMass = new Vector();
        private Vector clusterRadii = new Vector();
        private Vector clusterRmsDisplacement = new Vector();

        float MAX_COLLISION_DIST = 12f * (float) Math.pow(10.0, -9.0); // should be twice the radius of the largest particle in the simulation.
        float MAX_BOND_DIST = 30f * (float) Math.pow(10.0, -9.0);
        float PROBABILITY_OF_BOND_UPDATE = 0.4f;

        public Evolver(SimulationModel sm)
        {
                simulation = sm;
                timestep = 1f;
                monteCarloInit();
        }

        public void setTimestep(float ts)
        {
                timestep = ts;
                monteCarloInit();
        }

        public float getTimestep()
        {
                return timestep;
        }

        public void update()
        {
                monteCarloUpdate();
        }

        public void monteCarloInit()
        {
                waterViscosity = 0.001f; // pascal seconds.
                acceptanceFactor = 0.0f;
                acceptedChanges = 0;
                temperature = 300f;

                clusterObjects();
        }

        // perform one interation of the update routine
        public void monteCarloUpdate()
        {
                //long startTime = (new Date()).getTime();
                if (clusteredParticles.size() == 0) { clusterObjects(); }
                //int clusterNo = rand.nextInt(clusteredParticles.size());
                for (int i=0; i<clusteredParticles.size(); i++)
                {
                        moveCluster(i);
                }

                //long endMoveTime = (new Date()).getTime();
                //Vector cluster = (Vector)(clusteredIds.elementAt(clusterNo));
            	//String[] particleIds = (String[]) (clusteredParticles.elementAt(clusterNo));

                String[] allParticles = simulation.getAllParticles();
                for (int i=0; i<allParticles.length; i++)
                {
                        if (rand.nextFloat() > PROBABILITY_OF_BOND_UPDATE)
                        {
                                continue;
                        }

                        Vector3f particleCOM = simulation.getParticleCentreOfMass(allParticles[i]);

                        // Find the nearby particles
                        Vector nearbyParticles = simulation.getPDStore().getNearbyParticles(particleCOM, MAX_BOND_DIST);
                        Vector unformedBonds = updateParticleBonds(allParticles[i], nearbyParticles);
                        formNearbyBonds(unformedBonds);
                }
                //long endUpdateTime = (new Date()).getTime();

                //System.out.println("move   : "+(endMoveTime-startTime)+" ms");
                //System.out.println("update : "+(endUpdateTime-endMoveTime)+" ms");
        }

        public Vector updateParticleBonds(String particleId, Vector nearbyParticles)
        {
        	// test for bond changes
                Vector allPossibleBonds = new Vector();
                testBondChange(particleId, nearbyParticles, allPossibleBonds);
                clusterObjects();

                Vector addedBonds = new Vector();
                for (int i=0; i<allPossibleBonds.size(); i++)
                {
                        String[] elem = ((String[]) allPossibleBonds.elementAt(i));
                        if (elem[2].equals("all"))
                        {
                                addedBonds.addElement(elem);
                        }
                }

                // apply a random bond from the list of possible bonds.
                int i;
                if (addedBonds.size() == 0)
                {
                        return allPossibleBonds;
                }
                else if (addedBonds.size() == 1)
                {
                        i = 0;
                }
                else
                {
                        i = rand.nextInt(addedBonds.size()-1);
                }

                // get the id of the bonds and particles involved
                String[] newBondIds = (String[]) (addedBonds.elementAt(i));
                Bond newBond1 = (Bond)(simulation.getInstance(newBondIds[0]));
                Bond newBond2 = (Bond)(simulation.getInstance(newBondIds[1]));
                String partId1 = newBond1.getParentParticle();
                String partId2 = newBond2.getParentParticle();
                Particle part1 = (Particle)(simulation.getInstance(partId1));
                Particle part2 = (Particle)(simulation.getInstance(partId2));

                // get the bond points for each particle
                Vector3f[] bondPoints1 = part1.getBondPoints(newBondIds[0]);
                Vector3f[] bondPoints2 = part2.getBondPoints(newBondIds[1]);

                // get the cluster number of the object containing the particle being bonded to
                int clusterNo1 = 0;
                int partIndex1 = -1;
                int clusterNo2 = 0;
                int partIndex2 = -1;
                for (int j=0; j<clusteredParticles.size(); j++)
                {
                        String[] particlesToTest = (String[])(clusteredParticles.elementAt(j));
                        for (int k=0; k<particlesToTest.length; k++)
                        {
                                if (partId1.equals(particlesToTest[k]))
                                {
                                        partIndex1 = k;
                                        clusterNo1 = j;
                                }
                                if (partId2.equals(particlesToTest[k]))
                                {
                                        partIndex2 = k;
                                        clusterNo2 = j;
                                }
                        }
                        if (partIndex1 != -1 && partIndex2 != -1)
                        {
                                break;
                        }
                }
                if (partIndex1 == -1 || partIndex2 == -1 || clusterNo1 == clusterNo2)
                {
                        // the particle has not been found, or is on the same object as the other bond
                        return allPossibleBonds;
                }

                String[] particleIds1 = (String[]) (clusteredParticles.elementAt(clusterNo1));
                String[] particleIds2 = (String[]) (clusteredParticles.elementAt(clusterNo2));

                // calculate the centres of mass for the two objects
                Vector3f objectCOM1 = getCentreOfMassStored((Vector)(clusteredIds.elementAt(clusterNo1)));
                Vector3f objectCOM2 = getCentreOfMassStored((Vector)(clusteredIds.elementAt(clusterNo2)));

                // calculate the rotational difference between the bond points' normals
                // -- calculate the normal vectors for each bond point
                Vector3f normal1 = simulation.getVectorBetween(bondPoints1[0], bondPoints1[1]);
                Vector3f normal2 = simulation.getVectorBetween(bondPoints2[0], bondPoints2[1]);
                if (!Functions.normalize(normal1) || !Functions.normalize(normal2))
                {
                        return allPossibleBonds;
                }

                // invert normal2 (the normals should point in opposite directions)
                normal2.negate();

                Vector3f normalAxis = new Vector3f();
                float normalAngle = Functions.getAxisAndAngle(normalAxis, normal1, normal2);

                // apply the landscapes to determine whether the objects can rotate into normal alignment
                AxisAngle4f normalContrib1 = new AxisAngle4f(normalAxis, normalAngle);
                AxisAngle4f normalContrib2 = new AxisAngle4f(normalAxis, -normalAngle);
                for (int j=0; j<particleIds1.length; j++)
                {
                        Landscape partLandscape = simulation.getParticleLandscape(particleIds1[j]);
                        partLandscape.adjustRotation(normalContrib1, bondPoints1[0]);
                }
                for (int j=0; j<particleIds2.length; j++)
                {
                        Landscape partLandscape = simulation.getParticleLandscape(particleIds2[j]);
                        partLandscape.adjustRotation(normalContrib2, bondPoints2[0]);
                }

                // if the allowed normal rotations do not add up to the rotational difference, discard bond
                if ( (normalAngle - normalContrib1.angle + normalContrib2.angle > 0.001f))
                {
                        // discard
                return allPossibleBonds;
                }

                // if not discarded, calculate rotations of the two objects
                AxisAngle4f normalRot1 = new AxisAngle4f(normalAxis, 0f);
                AxisAngle4f normalRot2 = new AxisAngle4f(normalAxis, 0f);
                if ((normalContrib1.angle - normalContrib2.angle) > (float) Math.pow(10.0, -10.0) ||
                        (normalContrib1.angle - normalContrib2.angle) < -(float) Math.pow(10.0, -10.0))
                { normalRot1.angle =
                                normalContrib1.angle * normalAngle / (normalContrib1.angle - normalContrib2.angle);
                normalRot2.angle =
                                normalContrib2.angle * normalAngle / (normalContrib1.angle - normalContrib2.angle);}
                else
                { normalRot1.angle = 0f;
                normalRot2.angle = 0f; }

                // rotate the bond points
                Vector3f[] rotatedBondPoints1 =
                        Functions.transformPoints(bondPoints1, objectCOM1, new Vector3f(), normalRot1);
                Vector3f[] rotatedBondPoints2 =
                        Functions.transformPoints(bondPoints2, objectCOM2, new Vector3f(), normalRot2);

                // calculate the rotational difference between the bond points planar vectors
                // -- calculate the planar vectors
                Vector3f planar1 =  simulation.getVectorBetween(rotatedBondPoints1[0], rotatedBondPoints1[2]);
                Vector3f planar2 =  simulation.getVectorBetween(rotatedBondPoints2[0], rotatedBondPoints2[2]);
                if (!Functions.normalize(planar1) || !Functions.normalize(planar2))
                {
                        return allPossibleBonds;
                }

                Vector3f planarAxis = new Vector3f();
                float planarAngle = Functions.getAxisAndAngle(planarAxis, planar1, planar2);

                // apply the landscapes to determine whether the objects can rotate into planar alignment
                AxisAngle4f planarContrib1 = new AxisAngle4f(planarAxis, planarAngle);
                AxisAngle4f planarContrib2 = new AxisAngle4f(planarAxis, -planarAngle);
                for (int j=0; j<particleIds1.length; j++)
                {
                        Landscape partLandscape = simulation.getParticleLandscape(particleIds1[j]);
                        partLandscape.adjustRotation(planarContrib1, rotatedBondPoints1[0]);
                }
                for (int j=0; j<particleIds2.length; j++)
                {
                        Landscape partLandscape = simulation.getParticleLandscape(particleIds2[j]);
                        partLandscape.adjustRotation(planarContrib2, rotatedBondPoints2[0]);
                }

                // if the allowed normal rotations do not add up to the rotational difference, discard bond
                if ( (planarAngle - planarContrib1.angle + planarContrib2.angle > 0.001f))
                {
                        // discard
                        return allPossibleBonds;
                }

                // if not discarded, calculate the planar rotations of the two objects
                AxisAngle4f planarRot1 = new AxisAngle4f(planarAxis, 0f);
                AxisAngle4f planarRot2 = new AxisAngle4f(planarAxis, 0f);
                if ((planarContrib1.angle - planarContrib2.angle) > (float) Math.pow(10.0, -10.0) ||
                        (planarContrib1.angle - planarContrib2.angle) < -(float) Math.pow(10.0, -10.0))
                { planarRot1.angle =
                                planarContrib1.angle * planarAngle / (planarContrib1.angle - planarContrib2.angle);
                planarRot2.angle =
                                planarContrib2.angle * planarAngle / (planarContrib1.angle - planarContrib2.angle);}
                else
                { planarRot1.angle = 0f;
                planarRot2.angle = 0f; }

                // rotate the bond points by the planar rotation
                Vector3f[] planarRotatedBondPoints1 =
                        Functions.transformPoints(rotatedBondPoints1, objectCOM1, new Vector3f(), planarRot1);
                Vector3f[] planarRotatedBondPoints2 =
                        Functions.transformPoints(rotatedBondPoints2, objectCOM2, new Vector3f(), planarRot2);

                // calculate the translational difference between the objects
                Vector3f transDiff = simulation.getVectorBetween(planarRotatedBondPoints1[0],
                                planarRotatedBondPoints2[0]);

                // calculate the contribution to the translation made by each particle
                Vector3f contrib1 = new Vector3f(transDiff.x, transDiff.y, transDiff.z);
                Vector3f contrib2 = new Vector3f(-transDiff.x, -transDiff.y, -transDiff.z);
                for (int j=0; j<particleIds1.length; j++)
                {
                        Landscape partLandscape = simulation.getParticleLandscape(particleIds1[j]);
                        partLandscape.adjustDisplacement(contrib1, planarRotatedBondPoints1[0]);
                }
                for (int j=0; j<particleIds2.length; j++)
                {
                        Landscape partLandscape = simulation.getParticleLandscape(particleIds2[j]);
                        partLandscape.adjustDisplacement(contrib2, planarRotatedBondPoints2[0]);
                }

                // if not discarded, calculate displacements of the centres of mass of the objects.
                Vector3f trans1 = new Vector3f();
                Vector3f trans2 = new Vector3f();
                if ((contrib1.x - contrib2.x) > (float) Math.pow(10.0, -10.0) ||
                        (contrib1.x - contrib2.x) < -(float) Math.pow(10.0, -10.0))
                { trans1.x = contrib1.x * transDiff.x / (contrib1.x - contrib2.x);
                trans2.x = contrib2.x * transDiff.x / (contrib1.x - contrib2.x);}
                else
                { trans1.x = 0f;
                trans2.x = 0f; }
                if ((contrib1.y - contrib2.y) > (float) Math.pow(10.0, -10.0) ||
                        (contrib1.y - contrib2.y) < -(float) Math.pow(10.0, -10.0))
                { trans1.y = contrib1.y * transDiff.y / (contrib1.y - contrib2.y);
                trans2.y = contrib2.y * transDiff.y / (contrib1.y - contrib2.y); }
                else
                { trans1.y = 0f;
                trans2.y = 0f; }
                if ((contrib1.z - contrib2.z) > (float) Math.pow(10.0, -10.0) ||
                        (contrib1.z - contrib2.z) < -(float) Math.pow(10.0, -10.0))
                { trans1.z = contrib1.z * transDiff.z / (contrib1.z - contrib2.z);
                trans2.z = contrib2.z * transDiff.z / (contrib1.z - contrib2.z); }
                else
                { trans1.z = 0f;
                trans2.z = 0f; }

                // translate the bond points
                Vector3f[] translatedBondPoints1 =
                        Functions.transformPoints(planarRotatedBondPoints1, objectCOM1, trans1, new AxisAngle4f());
                Vector3f[] translatedBondPoints2 =
                        Functions.transformPoints(planarRotatedBondPoints2, objectCOM2, trans2, new AxisAngle4f());

                if ((new Float(normalRot1.x)).isNaN() ||
                (new Float(normalRot1.y)).isNaN() ||
                        (new Float(normalRot1.z)).isNaN())
                { return addedBonds; }
                if ((new Float(planarRot1.x)).isNaN() ||
                (new Float(planarRot1.y)).isNaN() ||
                        (new Float(planarRot1.z)).isNaN())
                { return addedBonds; }
                if ((new Float(normalRot2.x)).isNaN() ||
                (new Float(normalRot2.y)).isNaN() ||
                        (new Float(normalRot2.z)).isNaN())
                { return addedBonds; }
                if ((new Float(planarRot2.x)).isNaN() ||
                (new Float(planarRot2.y)).isNaN() ||
                        (new Float(planarRot2.z)).isNaN())
                { return addedBonds; }



                Vector3f[] particleComs1 = new Vector3f[particleIds1.length];
                for (int j=0; j<particleIds1.length; j++)
                {
                        particleComs1[j] = simulation.getParticleCentreOfMass(particleIds1[j]);
                }

                // apply the rotation
                Vector3f[] finalPos1 = Functions.transformPoints(particleComs1, objectCOM1, new Vector3f(), normalRot1);
                finalPos1 = Functions.transformPoints(finalPos1, objectCOM1, new Vector3f(), planarRot1);
                finalPos1 = Functions.transformPoints(finalPos1, objectCOM1, trans1, new AxisAngle4f());


                // -- test for collisions and boundary conditions
                Vector3f newCOM1 = new Vector3f();
                newCOM1.add(objectCOM1, trans1);
                boolean success1 = testUpdate(particleIds1, particleIds1, finalPos1, nearbyParticles);


                Vector3f[] particleComs2 = new Vector3f[particleIds2.length];
                for (int j=0; j<particleIds2.length; j++)
                {
                        particleComs2[j] = simulation.getParticleCentreOfMass(particleIds2[j]);
                }

                // apply the rotation
                Vector3f[] finalPos2 = Functions.transformPoints(particleComs2, objectCOM2, new Vector3f(), normalRot2);
                finalPos2 = Functions.transformPoints(finalPos2, objectCOM2, new Vector3f(), planarRot2);
                finalPos2 = Functions.transformPoints(finalPos2, objectCOM2, trans2, new AxisAngle4f());

                // -- test for collisions and boundary conditions
                Vector3f newCOM2 = new Vector3f();
                newCOM2.add(objectCOM2, trans2);
                boolean success2 = testUpdate(particleIds2, particleIds2, finalPos2, nearbyParticles);

                if (success1 && success2)
                {
                        // check for collisions between the two objects
                        if (objectsCollide(particleIds1, particleIds2, finalPos1, finalPos2))
                        {
                                return allPossibleBonds;
                        }

                        // set the new positions and bond points, if no collisions.
                        for (int j=0; j<particleIds1.length; j++)
                        {
                                simulation.rotateParticle(particleIds1[j], objectCOM1, normalRot1);
                                simulation.rotateParticle(particleIds1[j], objectCOM1, planarRot1);
                                simulation.translateParticle(particleIds1[j], trans1);
                                simulation.setParticleCentreOfMass(particleIds1[j], finalPos1[j]);

                        }
                        for (int j=0; j<particleIds2.length; j++)
                        {
                                simulation.rotateParticle(particleIds2[j], objectCOM2, normalRot2);
                                simulation.rotateParticle(particleIds2[j], objectCOM2, planarRot2);
                                simulation.translateParticle(particleIds2[j], trans2);
                                simulation.setParticleCentreOfMass(particleIds2[j], finalPos2[j]);
                        }

                        // set the bond
                        newBond1.bondTo(newBondIds[1]);
                        clusterObjects();

                        allPossibleBonds.removeElement(newBondIds);
                }
                return allPossibleBonds;
        }

        public void formNearbyBonds(Vector unformedBonds)
        {
                for (int i=0; i<unformedBonds.size(); i++)
                {
                        // get the id of the bonds and particles involved
                        String[] newBondIds = (String[]) (unformedBonds.elementAt(i));
                        Bond newBond1 = (Bond)(simulation.getInstance(newBondIds[0]));
                        Bond newBond2 = (Bond)(simulation.getInstance(newBondIds[1]));
                        String partId1 = newBond1.getParentParticle();
                        String partId2 = newBond2.getParentParticle();
                        Particle part1 = (Particle)(simulation.getInstance(partId1));
                        Particle part2 = (Particle)(simulation.getInstance(partId2));

                        // get the bond points for each particle
                        Vector3f[] bondPoints1 = part1.getBondPoints(newBondIds[0]);
                        Vector3f[] bondPoints2 = part2.getBondPoints(newBondIds[1]);

                        // get the cluster number of the object containing the particle being bonded to
                        int clusterNo1 = 0;
                        int partIndex1 = -1;
                        int clusterNo2 = 0;
                        int partIndex2 = -1;
                        for (int j=0; j<clusteredParticles.size(); j++)
                        {
                                String[] particlesToTest = (String[])(clusteredParticles.elementAt(j));
                                for (int k=0; k<particlesToTest.length; k++)
                                {
                                        if (partId1.equals(particlesToTest[k]))
                                        {
                                                partIndex1 = k;
                                                clusterNo1 = j;
                                        }
                                        if (partId2.equals(particlesToTest[k]))
                                        {
                                                partIndex2 = k;
                                                clusterNo2 = j;
                                        }
                                }
                                if (partIndex1 != -1 && partIndex2 != -1)
                                {
                                        break;
                                }
                        }
                        if (partIndex1 == -1 || partIndex2 == -1 || clusterNo1 != clusterNo2)
                        {
                                // reject since the particles have not been found, or are not on the same object cluster
                                break;
                        }

                        // reject if the particles are part of the same object
                        String objectId1= part1.getParentObjectId();
                        String objectId2 = part2.getParentObjectId();
                        if (objectId1.equals(objectId2))
                        {
                                break;
                        }

                        // reject if the bond sites are not close enough.
                        Vector3f bondSite1 = bondPoints1[0];
                        Vector3f bondSite2 = bondPoints2[0];
                        if ((float) Math.abs(bondSite1.x - bondSite2.x) > (float) Math.pow(10, -10))
                        { break;  }
                        if ((float) Math.abs(bondSite1.y - bondSite2.y) > (float) Math.pow(10, -10))
                        { break;  }
                        if ((float) Math.abs(bondSite1.z - bondSite2.z) > (float) Math.pow(10, -10))
                        { break;  }

                        // form the bond
                        newBond1.bondTo(newBondIds[1]);
                }
        }

        public boolean objectsCollide(String[] parts1, String[] parts2, Vector3f[] finalPos1, Vector3f[] finalPos2)
        {
                float radSum, dist;
                Vector3f partCOM1, partCOM2;
                Vector3f newPos1, newPos2;
                Particle particle1, particle2;

                for (int i=0; i<parts1.length; i++)
                {
                        for (int j=0; j<parts2.length; j++)
                        {
                                dist = simulation.getSquaredDistance(finalPos1[i], finalPos2[j]);
                                if (dist < MAX_COLLISION_DIST)
                                {
                                        radSum = simulation.getParticleRadius(parts1[i]) + simulation.getParticleRadius(parts2[j]);
                                        radSum *= radSum;
                                        if (dist < 0.98 * radSum)
                                        {
                                                 return true;
                                        }
                                }
                        }

                }
                return false;
        }

        public void moveCluster(int clusterNo)
        {
                /////////////////////////////////////////////////////////////////////////////
                // -- Calculate a possible new position and orientation for the cluster -- //
                /////////////////////////////////////////////////////////////////////////////
                String[] particleIds = (String[]) (clusteredParticles.elementAt(clusterNo));

                Vector cluster = (Vector)(clusteredIds.elementAt(clusterNo));
                calculateClusterComAndRadius(cluster);
                float objectRadius = getRadiusStored(cluster);
                Vector3f objectCOM = getCentreOfMassStored(cluster);

                // -- calculate a possible new position of the object
                float rmsDisplacement = getRmsDisplacementStored(cluster);
                float xDisp = rmsDisplacement * (float) (rand.nextGaussian());
                float yDisp = rmsDisplacement * (float) (rand.nextGaussian());
                float zDisp = rmsDisplacement * (float) (rand.nextGaussian());
                Vector3f displacement = new Vector3f(xDisp, yDisp, zDisp);

                //displacement = new Vector3f();

                // calculate a rotation for the object
                Vector3f objectRotationAxis;
                do
                {
                        objectRotationAxis = new Vector3f(rand.nextFloat() - 0.5f,
                                                        rand.nextFloat() - 0.5f,
                                                        rand.nextFloat() - 0.5f);

                        for (int i=0; i<particleIds.length; i++)
                        {
                                Landscape landscape = simulation.getParticleLandscape(particleIds[i]);
                                landscape.adjustDiffusionAxis(objectRotationAxis, objectCOM);
                        }
                 } while (!Functions.normalize(objectRotationAxis));

                // allow the rotational displacement along the equivalent circumference to be equal to the rms displacement
                float rotFactor = 360f * rmsDisplacement / (2f * (float) Math.PI * objectRadius);
                AxisAngle4f objectRotation =
                                        new AxisAngle4f(objectRotationAxis,
                                                                rotFactor * (float) (rand.nextGaussian()) * (float) Math.PI / 180f);

                // -- calculate the effect of the displacement on the object's centre of mass
                Vector3f newObjectCOM = new Vector3f(displacement.x+objectCOM.x, displacement.y+objectCOM.y, displacement.z+objectCOM.z);

                Vector3f[] finalPos = calculateNewCOMs(particleIds, objectCOM, displacement, objectRotation);
                ////////////////////////////////////////////////////////////
                // -- Calculate the new and old energies of the object -- //
                ////////////////////////////////////////////////////////////

                float oldEnergy = 0f;
                float newEnergy = 0f;
                for (int i=0; i<particleIds.length; i++)
                {
                        Landscape landscape = simulation.getParticleLandscape(particleIds[i]);
                        oldEnergy += landscape.valueAt(simulation.getParticleCentreOfMass(particleIds[i]));
                        newEnergy += landscape.valueAt(finalPos[i]);
                }
			
                totalChanges++;

                /////////////////////////////////////////////////////
                // -- First-level Monte Carlo simulation update -- //
                /////////////////////////////////////////////////////

                // -- do not update unless the new energy is lower than the old energy,
			// -- or the exponential is less than a random amount
                if ((newEnergy <= oldEnergy) ||
			((float) Math.exp(-(newEnergy - oldEnergy) / oldEnergy)) > rand.nextFloat() + acceptanceFactor)
                {
                        // -- get a list of other particles within collision distance of this object
                        boolean success = true;
                        for (int i=0; i<particleIds.length; i++)
                        {
                               Vector particlesToCompare = simulation.getPDStore().getNearbyParticles(
                                                                       finalPos[i], MAX_COLLISION_DIST);

                                // -- test for collisions and boundary conditions
                                String[] collideParticle = new String[1];
                                collideParticle[0] = particleIds[i];
                                success &= testUpdate(particleIds, collideParticle, finalPos, particlesToCompare);
                        }
					
                        if (success)
                        {
                    	
                                boolean validRotation = true;

                                AxisAngle4f[] newOrientations = new AxisAngle4f[particleIds.length];
                                for (int i=0; i<particleIds.length; i++)
                                {
                                        AxisAngle4f partRotation = simulation.getParticleOrientation(particleIds[i]);

                                        newOrientations[i] = Functions.combine(partRotation, objectRotation);
                                        if (newOrientations[i] == null)
                                        {
                                                validRotation = false;
                                        }
                                }
                    		 
                    		 
                                if (validRotation)
                                {
                                        // rotate the individual particles and set the new positions.
                                        for (int i=0; i<particleIds.length; i++)
                                        {

                                                simulation.setParticleOrientation(particleIds[i], newOrientations[i]);
                                                simulation.setParticleCentreOfMass(particleIds[i], finalPos[i]);

                                                // -- update the positions of the particle's bonds
                                                Vector3f[] oldBondPoints = simulation.getParticleBondPoints(particleIds[i]);
                                                Vector3f[] newBondPoints = Functions.transformPoints(oldBondPoints, objectCOM,
                                                                                                displacement, objectRotation);

                                                simulation.setParticleBondPoints(particleIds[i], newBondPoints);
                                        }
                                        displaceCentreOfMass(cluster, displacement);
                        	}
                        }
                }
        }

        public Vector3f[] calculateNewCOMs(String[] particleIds, Vector3f objectCOM,
                                                                        Vector3f displacement, AxisAngle4f objectRotation)
        {
                // adjust the rotation by the particle landscape restrictions
                for (int i=0; i<particleIds.length; i++)
                {
                        Landscape partLandscape = simulation.getParticleLandscape(particleIds[i]);
                        partLandscape.adjustRotation(objectRotation,
                                        simulation.getParticleCentreOfMass(particleIds[i]));
                }

                Vector3f[] particleComs = new Vector3f[particleIds.length];
                for (int i=0; i<particleIds.length; i++)
                {
                        particleComs[i] = simulation.getParticleCentreOfMass(particleIds[i]);
                }

                // apply the rotation
                Vector3f[] rotatedParticles =
                                        Functions.transformPoints(particleComs, objectCOM, new Vector3f(), objectRotation);

                // adjust the displacement by the particle landscape restrictions
                for (int i=0; i<particleIds.length; i++)
                {
                        Landscape partLandscape = simulation.getParticleLandscape(particleIds[i]);
                        partLandscape.adjustDisplacement(displacement, rotatedParticles[i]);
                }

                Vector3f[] finalPos =
                                Functions.transformPoints(rotatedParticles, objectCOM, displacement, new AxisAngle4f());

                return finalPos;
        }

        public void clusterObjects()
        {
                // -- cluster the objects according to the bonds
                String[] objectIds = simulation.getRootObjects();
                Vector objectIdVect = new Vector();
                Vector objectParticleIds = new Vector();
                Vector objectIdCluster = new Vector();
                for (int i=0; i<objectIds.length; i++)
                {
                        objectIdVect.addElement(objectIds[i]);
                        objectIdCluster.addElement(new Vector());
                        ((Vector)(objectIdCluster.elementAt(i))).addElement(objectIds[i]);
                        objectParticleIds.addElement(simulation.getAllParticlesOfObject(objectIds[i]));
                }

                for (int i=0; i<objectParticleIds.size(); i++)
                {
                        String[] thisObjectsParticles = (String[]) (objectParticleIds.elementAt(i));
                        for (int j=0; j<thisObjectsParticles.length; j++)
                        {
                                Particle thisParticle = (Particle) (simulation.getInstance(thisObjectsParticles[j]));
                                Vector thisParticlesBondIds = thisParticle.getBondIds();
                                for (int k=0; k<thisParticlesBondIds.size(); k++)
                                {
                                        Bond thisBond =
                                                (Bond) (simulation.getInstance((String) (thisParticlesBondIds.elementAt(k))));
                                        String connectedBondId = thisBond.getConnectedBond();
                                        if (connectedBondId != "")
                                        {
                                                Bond connectedBond = (Bond) (simulation.getInstance(connectedBondId));
                                                Particle connectedParticle =
                                                                (Particle) (simulation.getInstance(connectedBond.getParentParticle()));
                                                String connectedObjectId = connectedParticle.getParentObjectId();
                                                int objectIndex = objectIdVect.indexOf(connectedObjectId);
                                                // -- only cluster the bonded objects if they are not the same object
                                                if (objectIndex != i)
                                                {
                                                        String[] connectedObjectParticles =
                                                                        (String[]) (objectParticleIds.elementAt(objectIndex));
                                                        String[] newCluster =
                                                                new String[thisObjectsParticles.length + connectedObjectParticles.length];

                                                        for (int m=0; m<thisObjectsParticles.length; m++)
                                                        {
                                                                        newCluster[m] = thisObjectsParticles[m];
                                                        }
                                                        for (int m=0; m<connectedObjectParticles.length; m++)
                                                        {
                                                                        newCluster[m+thisObjectsParticles.length] = connectedObjectParticles[m];
                                                        }

                                                        thisObjectsParticles = newCluster;
                                                        objectParticleIds.removeElementAt(i);
                                                        objectParticleIds.insertElementAt(newCluster,i);
                                                        objectParticleIds.removeElementAt(objectIndex);
                                                        objectParticleIds.insertElementAt(new String[0], objectIndex);

                                                        // -- cluster the ids
                                                        Vector idVector1 = (Vector)(objectIdCluster.elementAt(i));
                                                        Vector idVector2 = (Vector)(objectIdCluster.elementAt(objectIndex));
                                                        Vector newIdCluster = new Vector();
                                                        for (int m=0; m<idVector1.size(); m++)
                                                        {
                                                                newIdCluster.addElement(idVector1.elementAt(m));
                                                        }
                                                        for (int m=0; m<idVector2.size(); m++)
                                                        {
                                                                newIdCluster.addElement(idVector2.elementAt(m));
                                                        }

                                                        objectIdCluster.removeElementAt(i);
                                                        objectIdCluster.insertElementAt(newIdCluster, i);
                                                        objectIdCluster.removeElementAt(objectIndex);
                                                        objectIdCluster.insertElementAt(new Vector(), objectIndex);
                                                }
                                        }
                                }
                        }
                }

                // Compile a list of the clusters (removing the empty clusters)
                clusteredParticles = new Vector();
                clusteredIds = new Vector();
                for (int i=0; i<objectParticleIds.size(); i++)
                {
                        String[] objPartIds = (String[]) (objectParticleIds.elementAt(i));
                        if (objPartIds.length > 0)
                        {
                                clusteredParticles.addElement(objPartIds);
                                clusteredIds.addElement(objectIdCluster.elementAt(i));
                        }
                }

                // reset the centres of mass and radii
                clusterLookup = new Vector();
                clusterCentresOfMass = new Vector();
                clusterRadii = new Vector();
                clusterRmsDisplacement = new Vector();
        }

        public void displaceCentreOfMass(Object c, Vector3f displacement)
        {
                Vector cluster = (Vector) c;

                // check if the cluster is already in the list
                int index = clusterLookup.indexOf(cluster);
                if (index == -1)
                {
                        calculateClusterComAndRadius(cluster);
                }
                else
                {
                        Vector3f com = (Vector3f)(clusterCentresOfMass.elementAt(index));
                        com.x += displacement.x;
                        com.y += displacement.y;
                        com.z += displacement.z;
                }
                if (cluster.size() > 1)
                {
                        for (int i=0; i<cluster.size(); i++)
                        {
                                Vector singleObject = new Vector();
                                singleObject.addElement(cluster.elementAt(i));
                                displaceCentreOfMass(singleObject, displacement);
                        }
                }
        }

        public Vector3f getCentreOfMassStored(Object cluster)
        {
                // always calculate the centre of mass and radius
                calculateClusterComAndRadius(cluster);

                int index = clusterLookup.indexOf(cluster);
                if (index == -1)
                {
                        calculateClusterComAndRadius(cluster);
                        index = clusterLookup.indexOf(cluster);
                }
                Vector3f com = (Vector3f) (clusterCentresOfMass.elementAt(index));
                if (com == null)
                {
                        calculateClusterComAndRadius(cluster);
                        return (Vector3f) (clusterCentresOfMass.elementAt(index));
                }
                else
                {
                        return com;
                }
        }

        public float getRadiusStored(Object cluster)
        {
                int index = clusterLookup.indexOf(cluster);
                if (index == -1)
                {
                        calculateClusterComAndRadius(cluster);
                        index = clusterLookup.indexOf(cluster);
                }
                Float rad = (Float) (clusterRadii.elementAt(index));
                if (rad == null)
                {
                        calculateClusterComAndRadius(cluster);
                        return ((Float) (clusterRadii.elementAt(index))).floatValue();
                }
                else
                {
                        return rad.floatValue();
                }
        }

        public void calculateClusterComAndRadius(Object c)
        {
                Vector cluster = (Vector) c;

                //String[] particleIds = (String[]) (clusteredParticles.elementAt(clusterNo));
                // get all particles in the object
                Vector particleIds = new Vector();
                for (int i=0; i<cluster.size(); i++)
                {
                        String objectId = (String) cluster.elementAt(i);
                        String[] particles = simulation.getAllParticlesOfObject(objectId);
                        for (int j=0; j<particles.length; j++)
                        {
                                particleIds.addElement(particles[j]);
                        }
                }

                // -- calculate a centre of mass and effective radius for the object
                Vector3f clusterCentreOfMass = new Vector3f();
                float objectVol = 0f;
                float totalMass = 0f;
                for (int i=0; i<particleIds.size(); i++)
                {
                        String thisId = (String)(particleIds.elementAt(i));
                        Vector3f particleCOM = simulation.getParticleCentreOfMass(thisId);
                        float particleMass = simulation.getParticleMass(thisId);
                        totalMass += particleMass;
                        clusterCentreOfMass.x += particleCOM.x * particleMass;
                        clusterCentreOfMass.y += particleCOM.y * particleMass;
                        clusterCentreOfMass.z += particleCOM.z * particleMass;

                        float particleRadius = simulation.getParticleRadius(thisId);
                        objectVol += simulation.getParticleVol(thisId);
                }
                clusterCentreOfMass.x /= totalMass;
                clusterCentreOfMass.y /= totalMass;
                clusterCentreOfMass.z /= totalMass;

                Float clusterRadius = new Float((float) Math.pow(3f * objectVol / ((float) Math.PI * 4f), (1f / 3f)) );

                int index = clusterLookup.indexOf(cluster);
                if (index == -1)
                {
                        clusterLookup.addElement(cluster);
                        clusterCentresOfMass.addElement(clusterCentreOfMass);
                        clusterRadii.addElement(clusterRadius);
                        clusterRmsDisplacement.addElement(null);
                }
                else
                {
                        clusterCentresOfMass.removeElementAt(index);
                        clusterCentresOfMass.insertElementAt(clusterCentreOfMass, index);
                        clusterRadii.removeElementAt(index);
                        clusterRadii.insertElementAt(clusterRadius, index);
                }
        }


        public float getRmsDisplacementStored(Object cluster)
        {
                int index = clusterLookup.indexOf(cluster);
                if (index == -1)
                {
                        calculateClusterRmsDisplacement(cluster);
                        index = clusterLookup.indexOf(cluster);
                }
                Float disp = (Float) (clusterRmsDisplacement.elementAt(index));
                if (disp == null)
                {
                        calculateClusterRmsDisplacement(cluster);
                        return ((Float) (clusterRmsDisplacement.elementAt(index))).floatValue();
                }
                else
                {
                        return disp.floatValue();
                }
        }

        public void calculateClusterRmsDisplacement(Object cluster)
        {
                // -- compute the root-mean-square displacement of the object
                float dragCoefficient = 6f * (float) Math.PI * waterViscosity * getRadiusStored(cluster);
                diffusionCoefficient = BOLTZMANN_CONSTANT * temperature / dragCoefficient; // Einstein relation
                Float clusterRmsDisp = new Float((float) Math.sqrt(2f * diffusionCoefficient * timestep));

                int index = clusterLookup.indexOf(cluster);
                if (index == -1)
                {
                        clusterLookup.addElement(cluster);
                        clusterCentresOfMass.addElement(null);
                        clusterRadii.addElement(null);
                        clusterRmsDisplacement.addElement(clusterRmsDisp);
                }
                else
                {
                        clusterRmsDisplacement.removeElementAt(index);
                        clusterRmsDisplacement.insertElementAt(clusterRmsDisp, index);
                }
        }

        public void testBondChange(String objParticle, Vector allParticles, Vector addedBonds)
        {
                // -- compare with all nearby particles and particles on the same object
                String[] particleArray = new String[allParticles.size()-1];
                int count=0;
                //for (int j=0; j<allParticles.size(); j++)
                for (int j=allParticles.size()-1; j>=0; j--)
                {
                        String particle = (String) (allParticles.elementAt(j));

                        if (!particle.equals(objParticle))
                        {
                                particleArray[count] = particle;
                                count++;
                        }
                }
                Particle p = (Particle)
                                (simulation.getInstance(objParticle));
                Vector bondIds = p.getBondIds();

                //for (int k=0; k<bondIds.size(); k++)
                for (int k=bondIds.size()-1; k>=0; k--)
                {
                        Bond bond = (Bond) (simulation.getInstance((String)(bondIds.elementAt(k))));
                        bond.update(particleArray, addedBonds);
                }
        }


        // -- tests to see if the new update position is legal (ie. no collisions, etc.)
        public boolean testUpdate(String[] objectParticleIds, String[] particleIds, Vector3f[] finalPos, Vector particlesToCompare)
        {
                boolean success;

                // -- check for collisions
                for (int i=0; i<objectParticleIds.length; i++)
                {
                        if (!(simulation.testPosition(finalPos[i]))) { return false; }
                }

                // -- check for inter-object collisions

                Vector collisionParticles = new Vector(particlesToCompare);

                // remove elements in the current object (these can collide)
                for (int i=0; i<objectParticleIds.length; i++)
                {
                        collisionParticles.removeElement(objectParticleIds[i]);
                }

                // remove static particles
                for (int i=0; i<collisionParticles.size(); i++)
                {
                        Particle p =
                                (Particle)(simulation.getInstance((String)(collisionParticles.elementAt(i))));
                        if (!(p.collide()))
                        {
                                collisionParticles.removeElementAt(i);
                        }
                }

                // check for collisions between the remaining particles and the objects particles
                for (int i=0; i<particleIds.length; i++)
                {
                        int index = 0;
                        for (int j=0; j<objectParticleIds.length; j++)
                        {
                                if (objectParticleIds[j].equals(particleIds[i]))
                                {
                                        index = j;
                                }
                        }

                        for (int j=0; j<collisionParticles.size(); j++)
                        {
                                String collisionId = (String) (collisionParticles.elementAt(j));
                                if (!collisionId.equals(""))
                                {
                                        // check the two particles are not closer than the sum of their radii
                                        Vector3f partCOM = finalPos[index];
                                        float partRad = simulation.getParticleRadius(particleIds[i]);
                                        Vector3f collCOM = simulation.getParticleCentreOfMass(collisionId);
                                        float collRad = simulation.getParticleRadius(collisionId);
                                        float radSum = partRad+collRad;
                                        if (radSum*radSum > simulation.getSquaredDistance(partCOM, collCOM))
                                        {
                                                return false;
                                        }
                                }
                        }
                }
                return true;
        }

}